import itertools
import threading
import time
import sys


class LoadingScreen():
    def __init__(self):
        self.done = False
        self.t = None

    def __animate(self, data_type):
        for c in itertools.cycle(['|', '/', '-', '\\']):
            if self.done:
                break

            sys.stdout.write(f'\rDownloading {data_type} data from KNMI' + c)
            sys.stdout.flush()
            time.sleep(0.2)

    def start(self, data_type):
        self.t = threading.Thread(target=self.__animate, args=[data_type])
        self.t.start()

    def stop(self):
        sys.stdout.flush()
        sys.stdout.write('\rSuccess\n')
        self.done = True




#!/usr/bin/env python3

__author__ = "Ugurcan Akpulat"
__copyright__ = "Copyright 2021, Eleena Software"
__credits__ = [""]
__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Ugurcan Akpulat"
__email__ = "ugurcan.akpulat@gmail.com"
__status__ = "Production"

import aiohttp
import asyncio
import time
import os
import shutil
import traceback
from io import BytesIO
from zipfile import ZipFile
from .loading_screen import LoadingScreen


class KNMIDataLoader():
    """This class downloads the hourly weather data
        from KNMI klimatologie page and saves it to csv files at
        given path. Those files are updated daily.
    """
    def __init__(self, sbegin: int, send: int, data_type: str, dir_name=None,
                 download=False):
        self.__type = data_type
        self.__sbegin = sbegin
        self.__send = send

        if not dir_name:
            if self.__type == 'hourly':
                dir_name = 'station_uur_temp'
            else:
                dir_name = 'station_daily_temp'

        _path = os.path.join(os.getcwd(), dir_name)
        self.__output_path = _path

        if os.path.exists(_path) and download:
            shutil.rmtree(_path)

        if not os.path.exists(_path):
            os.makedirs(_path)

    def manipulate_text(self, zip: ZipFile, contained_file: str,
                        file_path: str) -> str:
        text = zip.open(contained_file).read().decode()
        text = text[text.index('STN,'):].replace(" ", "")
        header = "TXH,PG,PX,PXH,PN,PNH,VVN,VVNH,VVX,VVXH,NG,UG,UX,UXH,UN,UNH"
        replace = "TXH,PG,PX,PXH,PN,PNH,VVN,VVNH,VVX,VVXH,NG,UG,UX,UXH,UN,UNH,"
        text = text.replace(header, replace)
        text = text.replace("\n,", ",", 1)
        if os.path.exists(file_path):
            text = text[text.find('\n')+3:]
        return text

    def unpack_zip(self, content: aiohttp.ClientResponse, i: int) -> bool:
        with ZipFile(BytesIO(content)) as zip:
            file = f'{i}'
            for contained_file in zip.namelist():
                file_path = (os.path.join(self.__output_path, file + ".csv"))
                text = self.manipulate_text(zip, contained_file, file_path)
                with open(file_path, "a") as output:
                    output.write(text)
            return True

    def generate_tasks(self, s):
        tasks = []
        for i in range(self.__sbegin, self.__send):
            if self.__type == 'hourly':
                for j in ['2001-2010', '2011-2020', '2021-2030']:
                    url = f'https://cdn.knmi.nl/knmi/map/page/klimatologie/gegevens/uurgegevens/uurgeg_{i}_{j}.zip'
                    task = asyncio.ensure_future(self.save_to_file(s, url, i))
                    tasks.append(task)
            else:
                url = f'https://cdn.knmi.nl/knmi/map/page/klimatologie/gegevens/daggegevens/etmgeg_{i}.zip'
                task = asyncio.ensure_future(self.save_to_file(s, url, i))
                tasks.append(task)

        return tasks

    async def save_to_file(self, session: aiohttp.ClientSession, url: str,
                           i: int) -> bool:
        async with session.get(url) as resp:
            try:
                content = await resp.read()
                try:
                    if 'Error' in content.decode("utf-8"):
                        pass
                except UnicodeDecodeError:
                    if len(content) > 0:
                        return self.unpack_zip(content, i)
            except Exception:
                traceback.print_exc()

    async def fetch(self) -> None:
        async with aiohttp.ClientSession() as session:
            tasks = self.generate_tasks(session)
            tasks = await asyncio.gather(*tasks)
            for task_result in tasks:
                pass

    async def start(self, data_type) -> None:
        try:
            ls = LoadingScreen()
            ls.start(data_type)
            start_time = time.time()
            await self.fetch()
            ls.stop()
            print("--- %s seconds ---" % (time.time() - start_time))
        except Exception:
            traceback.print_exc()



__author__ = "Ugurcan Akpulat"
__copyright__ = "Copyright 2021, Eleena Software"
__credits__ = [""]
__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Ugurcan Akpulat"
__email__ = "ugurcan.akpulat@gmail.com"
__status__ = "Production"

import pandas as pd
import os
import knmi
import asyncio
import sys
from .KNMI_data_fetch import KNMIDataLoader
from geopy.distance import geodesic
from geopy.geocoders import Nominatim


class KNMI():
    """
    Class responsible for loading csv files into memory
    """

    def __init__(self, data_type='hourly', download=False):
        dir_name = 'station_uur_temp'
        if data_type == 'hourly':
            loader = KNMIDataLoader(1, 1000, 'hourly', dir_name, download)
        elif data_type == 'daily':
            dir_name = 'station_dag_temp'
            loader = KNMIDataLoader(1, 1000, 'daily', dir_name, download)
        else:
            raise ValueError('Use hourly or daily for data_type')

        if download:
            asyncio.run(loader.start(data_type))

        self.__path = os.path.join(os.getcwd(), dir_name)
        self.__stations = {}
        self.__downloaded_station_numbers = [int(f.split('.')[0]) for f in
                                             os.listdir(self.__path)
                                             if f.endswith('.csv')]
        self.__relevant_station_numbers = []

        for nmbr in self.__downloaded_station_numbers:
            file_path = os.path.join(self.__path, str(nmbr) + '.csv')
            sys.stdout.write(f'\rLoading station: {nmbr}')
            dtypes = self.__dtypes(file_path)
            self.__stations[nmbr] = pd.read_csv(os.path.join(
                                                self.__path, file_path),
                                                dtype=dtypes)

        sys.stdout.write('\n')

    def __len__(self):
        return len(self.__stations)

    def __getitem__(self, position):
        return self.__stations[position]

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self):
            result = self[self.__downloaded_station_numbers[self.n]]
            self.n += 1
            return result
        else:
            raise StopIteration

    def __dtypes(self, path: str) -> dict:
        headers = pd.read_csv(path, index_col=0, nrows=0).columns.tolist()
        dtypes = {headers: float for headers in headers}
        dtypes['YYYYMMDD'] = str
        return dtypes

    def __find_closest_st(self, lat: int, lng: int) -> knmi.metadata.Station:
        stations = [knmi.stations.get(station)
                    for station in self.__downloaded_station_numbers]
        minDistance = geodesic((lat, lng), (stations[0].latitude,
                                            stations[0].longitude)).meters
        for station in stations:
            tempDistance = geodesic((lat, lng), (station.latitude,
                                                 station.longitude)).meters
            if tempDistance < minDistance:
                minDistance = tempDistance
                closestStation = station
        return closestStation

    def get_coordinates_from_postcode_NL(self, postalcode: str, cnt='NL'):
        geolocator = Nominatim(user_agent="knmi-app")
        location = geolocator.geocode(query={'postalcode': postalcode,
                                             'country': cnt})
        return location

    def find_closest_station_number(self, postcode: str) -> int:
        cord = self.get_coordinates_from_postcode_NL(postcode)
        station = self.__find_closest_st(cord.latitude, cord.longitude)
        return station.number

    def find_df(self, postcode: str) -> pd.DataFrame:
        station_number = self.find_closest_station_number(postcode)
        return self.__stations[station_number]

    def get_all_info(self):
        df = None
        for k in self:
            if df is None:
                df = k
            else:
                df.append(k)
        return df

    def eleena_daily_get(self):
        df = None
        for k in self:
            try:
                if df is None:
                    df = k[['STN', 'YYYYMMDD', 'TG', 'SQ', 'DR', 'NG', 'Q']]
                else:
                    df = df.append(k[['STN', 'YYYYMMDD', 'TG', 'SQ', 'DR', 'NG', 'Q']])

            except Exception:
                continue

        self.__relevant_station_numbers.append(df['STN'][0])
        return df

    def find_closest_station_number_relevant(self, lat, long):
        station = self.__find_closest_st(lat, long)
        return station.number
